import click
import pandas as pd
import geopandas as gpd
import json
import yaml

from . import sources
from . import connection

pd.set_option('display.max_columns', None)
pd.set_option("display.max_rows", None)
pd.set_option('display.expand_frame_repr', False)

@click.group()
def main():
    pass

@main.group()
def source():
    pass

@source.command()
@click.option('--long', is_flag=True, default=False)
# @click.option('--outformat', default="sgf", help='Ouput format: %s' % ", ".join(dumpers.keys()))
# @click.argument('input', type=str)
# @click.argument('output', type=str)
def list(long=False):
    s = sources.load()
    #s = s.join(s.geometry.bounds)
    s["bbox"] = s.geometry.bounds.apply(lambda row: "%(minx).4f,%(miny).4f,%(maxx).4f,%(maxy).4f" % row, axis=1)
    
    s["url"] = s.connection_args.apply(lambda a: a["url"])
    s = s.drop(columns=["connection_args", "geometry"])

    maincols = ["layer", "connection_type", "url", "crs_orig", "bbox"]
    if long:
        cols = maincols + sorted(set(s.columns) - set(maincols))
    else:
        cols = maincols

    print(s[cols])
    
@source.command()
@click.argument('title', type=str)
@click.argument('connection_type', type=str)
@click.argument('connection_args', type=str)
@click.argument('layer', type=str)
def add(**kw):
    kw["connection_args"] = json.loads(kw["connection_args"])
    sources.add_source(**kw)
    
@source.command()
@click.argument('yamlfile', type=str)
def add_mapproxy(yamlfile):
    with open(yamlfile) as f:
        data = yaml.load(f, Loader=yaml.Loader)
    sources.add_mapproxy(data)



from . import connection
import contextily
import xyzservices.lib
import numpy as np
import geopandas as gpd
import shapely.geometry
from rasterio.transform import Affine

class TileConnection(connection.Connection):
    bands = 3
    dtype = "uint8"
    
    def __init__(self, **kw):
        connection.Connection.__init__(self, **kw)

        args = {"name": "", "attribution": ""}
        args.update(self.kw["connection_args"])
        
        self.source = xyzservices.lib.TileProvider(**args)

    def get_shape(self):
        return gpd.GeoDataFrame(geometry=[shapely.geometry.box(*self.get_bounds())], crs=self.get_crs())
        
    def download(self, gdf, tif_res):
        gdf = gdf.to_crs(self.get_crs())
        xmin, ymin, xmax, ymax = gdf.total_bounds

        world_bounds = self.get_bounds()

        # From definition of spherical mercator:
        # tif_res = (world_bounds[2] - world_bounds[0]) / (256 * 2**zoom)
        zoom = int(np.ceil(np.log2((world_bounds[2] - world_bounds[0]) / (256 * tif_res))))

        array, extent = contextily.bounds2img(xmin, ymin, xmax, ymax, ll=False, zoom=zoom, source=self.source)
        array = np.transpose(array, (2, 0, 1))

        left, right, bottom, top = extent
        xres = (right - left) / array.shape[1]
        yres = (top - bottom) / array.shape[2]
        
        transform = Affine.translation(left, top) * Affine.scale(xres, -yres)
        data = dict(self.kw)
        data["crs_orig"] = self.get_crs()
        return {"array":array, "transform":transform, "data":data, "gdf":gdf}

    def download_tile(self, bounds, tif_res, size):        
        raise NotImplementedError("Use download()")

    def get_bounds(self):
        # Lon,lat bbox: (-180, -85.051129, 180, 85.051129)
        return (-20037508.342789244, -20037508.342789255, 20037508.342789244, 20037508.342789244)

    def get_crs(self):
        return 3857



import geopandas as gpd
import pandas as pd
import os.path
import pkg_resources
import traceback
from . import connection 

sources_path = os.path.expanduser("~/.config/terrainy/sources.geojson")

def load():
    with pkg_resources.resource_stream("terrainy", "sources.geojson") as f:
        sources = gpd.read_file(f).set_index("title")
    if os.path.exists(sources_path):
        with open(sources_path, "rb") as f:
            sources = pd.concat((
                sources,
                gpd.read_file(f).set_index("title"))
            )
    return sources.loc[~sources.index.duplicated(keep='first')]

def dump(sources):
    sources_dir = os.path.dirname(sources_path)
    if not os.path.exists(sources_dir):
        os.makedirs(sources_dir)
    sources.to_file(sources_path, driver='GeoJSON')

def add_source(**kw):
    con = connection.connect(**kw)
    kw["crs_orig"] = con.get_crs()
    kw["geometry"] = con.get_shape().to_crs(4326).iloc[0].geometry
    s = load()
    s.loc[kw.pop("title")] = kw
    dump(s)

def add_mapproxy(data):
    for title, spec in data["sources"].items():
        try:
            args = {"title": title, "connection_type": spec["type"], "connection_args": {}}
            if "url" in spec:
                args["connection_args"]["url"] = spec["url"]
            if "req" in spec:
                if "url" in spec["req"]:
                    args["connection_args"]["url"] = spec["req"]["url"]
                if "layers" in spec["req"]:
                    args["layer"] = spec["req"]["layers"]

            if "url" in args["connection_args"]:
                args["connection_args"]["url"] = args["connection_args"]["url"].replace("%(", "{").replace(")s", "}")

            if "grid" in spec:
                grid = spec["grid"]
                if grid == "GLOBAL_WEBMERCATOR":
                    args["crs_orig"] = "EPSG:3857"
                else:
                    args["crs_orig"] = data["grids"][grid]["srs"]
        except Exception as e:
            print("Parser error for source %s: %s" % (title, e))
            print("\n".join(("    " + line for line in traceback.format_exc().split("\n"))))
        else:
            try:
                add_source(**args)
            except Exception as e:
                print("Unable to add source %s: %s: %s" % (
                    title, args, e))
                print("\n".join(("    " + line for line in traceback.format_exc().split("\n"))))



from . import connection
from owslib.wms import WebMapService

class WmsConnection(connection.Connection):
    bands = 3
    dtype = "uint8"
    formats = ['image/geotiff',
               'image/png',
               'image/jpeg',
               'image/tiff']
    
    def __init__(self, **kw):
        connection.Connection.__init__(self, **kw)
        self.wms = WebMapService(**self.kw["connection_args"])
        self.layer = self.wms[self.kw["layer"]]

        supported = self.wms.getOperationByName('GetMap').formatOptions
        self.file_format = [fmt for fmt in self.formats if fmt in supported][0]

    def download_tile(self, bounds, tif_res, size):
        return self.wms.getmap(layers=[self.layer.id],
                               srs=self.get_crs(),
                               bbox=bounds,
                               size=size,
                               format=self.file_format)

    def get_bounds(self):
        return self.layer.boundingBox[:4]

    def get_crs(self):
        return self.layer.boundingBox[4]



from . import connection
from owslib.wcs import WebCoverageService

class WcsConnection(connection.Connection):
    bands = 1
    dtype = "float64"
    
    def __init__(self, **kw):
        connection.Connection.__init__(self, **kw)
        self.wcs = WebCoverageService(**self.kw["connection_args"])
        self.layer = self.wcs[self.kw["layer"]]

    def download_tile(self, bounds, tif_res, size):
        return self.wcs.getCoverage(
            identifier=self.layer.id,
            crs=self.get_crs(),
            bbox=bounds,
            resx=tif_res, resy=tif_res,
            format='GeoTIFF')

    def get_bounds(self):
        return self.layer.boundingboxes[0]["bbox"]
        
    def get_crs(self):
        return self.layer.boundingboxes[0]["nativeSrs"]



import rasterio
import rasterio.mask
from rasterio.transform import Affine
import rasterio
from rasterio import MemoryFile
from rasterio.plot import show
import rasterio.mask
from rasterio.transform import Affine
import rasterio.rio.clip
from rasterio.crs import CRS
import geopandas as gpd
import time
import numpy as np
from shapely.geometry import Polygon
from owslib.wcs import WebCoverageService
from owslib.wms import WebMapService
import pkg_resources
import shapely
import json

from . import connection
from . import sources
    
def download(gdf, title, tif_res):
    "Downloads raster data for a shape from a given source"
    data  = sources.load().loc[title]
    con = connection.connect(**data)
    return con.download(gdf, tif_res)

def clip_to_bounds(file, area):
    with rasterio.open(file) as src:
        bounds = shapely.geometry.box(**area.to_crs(src.crs).bounds.iloc[0].astype(int))
        out_image, out_transform = rasterio.mask.mask(src, [bounds], filled=False, crop=True)
        out_meta = src.meta.copy()
    out_meta.update({
        "driver": "GTiff",
        "height": out_image.shape[1],
        "width": out_image.shape[2],
        "transform": out_transform})
    with rasterio.open(file, "w", **out_meta) as dest:
        dest.write(out_image)
        
def export(data_dict, out_path):
    ras_meta = {'driver': 'GTiff',
                'dtype': data_dict["array"].dtype,
                'nodata': None,
                'width': data_dict["array"].shape[2],
                'height': data_dict["array"].shape[1],
                'count': data_dict["array"].shape[0],
                'crs': data_dict["data"]["crs_orig"],
                'transform': data_dict["transform"],
                'tiled': False,
                'interleave': 'band'}

    with rasterio.open(out_path, 'w', **ras_meta) as tif:
        tif.write(data_dict["array"])
    clip_to_bounds(out_path, data_dict["gdf"])


def get_maps(gdf):
    "Returns the available map sources available from your input shapefile"
    s = sources.load()
    s = s.loc[s.geometry.is_valid]
    return s.loc[s.contains(gdf["geometry"][0])]

def choose_map(title):
    "Returns the shape you want to use to get data from, based on the title"
    s = sources.load()
    return s.loc[s["title"]==title]

# Legacy names
getMaps = get_maps
chooseMap = choose_map
getDTM = download
getImagery = download
export_terrain = export
export_imagery = export

    
# fixme: Make clipping work to actual shape
# def getFeatures(gdf):
#     """Function to parse features from GeoDataFrame in such a manner that rasterio wants them, from
#     https: // automating - gis - processes.github.io / CSC18 / lessons / L6 / clipping - raster.html"""
#     import json
#     return [json.loads(gdf.to_json())['features'][0]['geometry']]
#
# def clipTif(raster, shape):
#     # with fiona.open(clip_shape_, "r") as shapefile:
#     #     shapes = [feature["geometry"] for feature in shapefile]
#     #shapes = shape.geometry[0]
#
#     with rasterio.open(raster) as src:
#         out_image, out_transform = rasterio.mask.mask(src, shape, crop=True)
#         out_meta = src.meta
#
#     out_meta.update({"driver": "GTiff",
#                      "height": out_image.shape[1],
#                      "width": out_image.shape[2],
#                      "transform": out_transform})











#!/usr/bin/env python

import setuptools
import os

setuptools.setup(
    name='terrainy',
    version='0.0.3',
    description='Auto-downloader for global terrain data and satellite imagery',
    long_description="""Library to download a raster of 
    global height data such as a DTM, or satellite imagery for a polygon.""",
    long_description_content_type="text/markdown",
    author='Ed Harrison',
    author_email='eh@emrld.no',
    url='https://github.com/emerald-geomodelling/terrainy',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'terrainy': ['*/*.geojson']},
    install_requires=[
        "rasterio",
        "geopandas",
        "shapely",
        "fiona",
        "owslib",
        "numpy",
        "rtree",
        "contextily",
	"xyzservices",
        "click"
    ],
    entry_points={
        'terrainy.connection': [
            'wcs = terrainy.connection_wcs:WcsConnection',
            'wms = terrainy.connection_wms:WmsConnection',
            'tile = terrainy.connection_tile:TileConnection',
        ],
        'console_scripts': [
            'terrainy = terrainy.cmd:main',
        ],
    },
)



import rasterio
import rasterio.mask
from rasterio.transform import Affine
import rasterio
from rasterio import MemoryFile
from rasterio.plot import show
import rasterio.mask
from rasterio.transform import Affine
import rasterio.rio.clip
from rasterio.crs import CRS
import geopandas as gpd
import time
import numpy as np
from shapely.geometry import Polygon
from owslib.wcs import WebCoverageService
from owslib.wms import WebMapService
import pkg_resources
import importlib.metadata
import shapely
import json
import contextlib

# Grid sizing
tile_pixel_length = 1024
tile_pixel_width = 1024

class Connection(object):
    def __init__(self, **kw):
        self.kw = kw

    def get_shape(self):
        bbox = self.get_bounds()
        empty_bbox = list(bbox)
        w = empty_bbox[2] - empty_bbox[0]
        empty_bbox[0] -= w
        empty_bbox[2] -= w
        with self.open_tile(empty_bbox,
                            (empty_bbox[2] - empty_bbox[0]) / tile_pixel_width,
                            (tile_pixel_width, tile_pixel_length)) as dataset:
            empty_data_array = dataset.read()
        
        with self.open_tile(bbox,
                            (bbox[2] - bbox[0]) / tile_pixel_width,
                            (tile_pixel_width, tile_pixel_length)) as dataset:
            data_array = dataset.read()

        xres = (bbox[2] - bbox[0]) / tile_pixel_width
        yres = (bbox[3] - bbox[1]) / tile_pixel_length
        transform = rasterio.transform.Affine.translation(bbox[0], bbox[3]) * rasterio.transform.Affine.scale(xres, -yres)            
        geometry = [shapely.geometry.shape(shp)
                    for shp, val in rasterio.features.shapes((data_array != empty_data_array).max(axis=0).astype("int16"), transform=transform)
                    if val > 0]
        if not len(geometry):
            raise ValueError("Map has only empty tiles!")
        
        return gpd.GeoDataFrame(
            geometry = [gpd.GeoDataFrame(geometry = geometry).geometry.unary_union]
        ).set_crs(self.get_crs())

    @contextlib.contextmanager
    def open_tile(self, *arg, **kw):
        response = self.download_tile(*arg, **kw)
        with MemoryFile(response) as memfile:
            with memfile.open() as dataset:
                yield dataset
    
    def download(self, gdf, tif_res):
        # Convert data back to crs of map
        gdf = gdf.to_crs(self.get_crs())
        xmin, ymin, xmax, ymax = gdf.total_bounds

        tile_m_length = tile_pixel_length * tif_res
        tile_m_width = tile_pixel_width * tif_res

        width = (xmax - xmin) / tif_res
        length = (ymax - ymin) / tif_res

        nr_cols = int(np.ceil(width / tile_pixel_length))
        nr_rows = int(np.ceil(length / tile_pixel_width))

        array = np.zeros((self.bands, tile_pixel_length * nr_rows, tile_pixel_width * nr_cols), dtype=self.dtype)

        for x_idx in range(nr_cols):
            for y_idx in range(nr_rows):
                print('Working on block %s,%s of %s,%s' % (x_idx + 1, y_idx + 1, nr_cols, nr_rows))

                x = xmin + x_idx * tile_m_width
                y = ymax - y_idx * tile_m_length - tile_m_length

                polygon = (Polygon(
                    [(x, y), (x + tile_m_width, y), (x + tile_m_width, y + tile_m_length), (x, y + tile_m_length)]))

                with self.open_tile(polygon.bounds, tif_res, (tile_pixel_width, tile_pixel_length)) as dataset:
                    data_array = dataset.read()

                    array[:, y_idx * tile_pixel_width:y_idx * tile_pixel_width + tile_pixel_width,
                          x_idx * tile_pixel_length:x_idx * tile_pixel_length + tile_pixel_length] = data_array[:, :, :]

        transform = Affine.translation(xmin, ymax) * Affine.scale(tif_res, -tif_res)
        return {"array":array, "transform":transform, "data":self.kw, "gdf":gdf}


def connect(**data):
    connections = {entry.name: entry.load()
               for entry in importlib.metadata.entry_points()['terrainy.connection']}
    if data["connection_type"] not in connections:
        raise NotImplementedError("Unknown connection type")
    return connections[data["connection_type"]](**data)
